/* Copyright 2003, Jan D. Wolter, All Rights Reserved. */

#include <ctype.h>

#include "backtalk.h"
#include "email.h"
#include "date.h"
#include "stack.h"
#include "sysdict.h"
#include "vstr.h"
#include "emailfunc.h"

#if EMAIL

/* These are the currently open emails.  They are in an array of size n_email
 * indexed by the email id.
 */
static Email **curr_email= NULL;
int n_email= 0;

/* BAD_ADDR - Check if an email address is in a sane form.  This is pretty
 * minimal right now.  We want printable characters, and one @ sign before
 * at least one dot.  Empty strings are OK.
 */

int bad_addr(char *addr)
{
    int at= FALSE, dot= FALSE;

    if (addr[0] == '\0') return FALSE;

    for (; *addr != '\0'; addr++)
    {
	if (!isascii(*addr) || !isprint(*addr))
	    return TRUE;
	if (at)
	{
	    if (*addr == '.') dot= TRUE;
	    if (*addr == '@') return TRUE;
	}
	else if (*addr == '@')
	    at= TRUE;
    }

    return !dot;
}


/* BAD_HEADER - Check if a string is legal header text - no newlines,
 * control characters or non-ascii stuff.  Empty strings are OK.
 */

int bad_header(char *str)
{
    int at= FALSE, dot= FALSE;

    for (; *str != '\0'; str++)
    {
	if (!isascii(*str) || !isprint(*str))
	    return TRUE;
    }
    return FALSE;
}


/* FUNC_EMAIL_NEW:   <subject> <reply-addr> <reply-name> <id> email_new <rc>
 *
 * Start a new email message and make it current email #<id>.  Any input may
 * be ().  Returns:
 *
 *   rc = 0     message created
 *   rc = 1     invalid subject
 *   rc = 2     invalid reply address
 *   rc = 3     invalid reply name
 */

void func_email_new()
{
    int rc;
    int emid= pop_integer();
    char *rep_name= pop_string();
    char *rep_addr= pop_string();
    char *subject=  pop_string();

    if (emid < 0 || emid > 255)
	die("Bad email id value %d",emid);

    if (bad_header(subject))
	rc= 1;
    else if (bad_addr(rep_addr))
	rc= 2;
    else if (bad_header(rep_name))
	rc= 3;
    else
    {
	/* Increase the size of the current email array enough so there is
	 * a slot for the new email.  Set all newly allocated slots to NULL
	 * pointers.
	 */
	if (curr_email == NULL)
	{
	    n_email= emid + 1;
	    curr_email= (Email **)calloc(n_email, sizeof(Email *));
	}
	else if (n_email <= emid)
	{
	    curr_email= (Email **)realloc(curr_email,(emid+1)*sizeof(Email *));
	    do {
		curr_email[n_email++]= NULL;
	    } while (n_email <= emid);
	}

	/* Deallocate an previous message with this id */
	if (curr_email[emid] != NULL)
	    email_free(curr_email[emid]);

	/* Create the new message */
	curr_email[emid]= new_email(NULL, NULL,
		(subject[0] == '\0') ? NULL : subject,
		(rep_addr[0] == '\0') ? NULL : rep_addr,
		(rep_name[0] == '\0') ? NULL : rep_name);

	rc= 0;
    }
    push_integer(rc);
    free(rep_name);
    free(rep_addr);
    free(subject);
}


/* FUNC_EMAIL_BODY:   <alt> <desc> <type> <text> <id> email_body <rc>
 *
 *  Attach a body to the current email message #<id>.  <alt> can be 0 for the
 *  primary version of the text, or 1 for the alternate.  If there is already
 *  a body with the same value of <alt>, then this is an attachment.  <desc>
 *  is a description.  It can be ().  Type is any mime/type.  If a message
 *  has a primary body of type (text/html) and no alternate body, then an
 *  alternate body of type (text/plain) is automatically generated. Returns:
 *
 *   rc = 0     text added
 *   rc = 1     invalid description
 *   rc = 2     invalid type
 *
 *  Function dies on invalid alt values or if there is no current email.
 */

void func_email_body()
{
    int rc;
    int emid= pop_integer();
    char *text= pop_string();
    char *type= pop_string();
    char *desc= pop_string();
    int alt= pop_integer();

    if (emid < 0 || emid >= n_email || curr_email[emid] == NULL)
	die("Email message #%d not defined",emid);
    else if (alt < 0 || alt > 1)
	die("Bad value for alt (%d)",alt);
    else if (bad_header(desc))
	rc= 1;
    else if (bad_header(type) || strchr(type,'/') == NULL)
	rc= 2;
    else
    {
	email_body(curr_email[emid], alt, type, ENC_DUNNO, text, -1,
		(desc[0] == '\0') ? NULL : desc);
	rc= 0;
    }
    push_integer(rc);
    free(text);
    free(type);
    free(desc);
}


/* FUNC_EMAIL_RECIP:  <type> <addr> <name> <id> email_recip <rc>
 *
 * Add a recipient to the current email message #<id>.  The recipient type may
 * be any of:
 *
 *   type = (T+)    append address to the "To" list
 *   type = (T)     set "To" list to contain only this address
 *   type = (C+)    append address to the "Cc" list
 *   type = (C)     set "Cc" list to contain only this address
 *   type = (B+)    append address to the "Bcc" list
 *   type = (B)     set "Bcc" list to contain only this address
 *
 * The address and name fields can be ().  Returns:
 *
 *   rc = 0     recipient added
 *   rc = 1     invalid address
 *   rc = 2     invalid name
 *
 *  Function dies on invalid type values or if there is no current email.
 */

void func_email_recip()
{
    int rc, recip;
    int emid= pop_integer();
    char *name= pop_string();
    char *addr= pop_string();
    char *type= pop_string();

    if (emid < 0 || emid >= n_email || curr_email[emid] == NULL)
	die("Email message #%d not defined",emid);

    if (isascii(type[0]) && islower(type[0]))
	type[0]= toupper(type[0]);

    recip= -1;
    switch (type[0])
    {
	case 'T':  recip= REC_TO;  break;
	case 'C':  recip= REC_CC;  break;
	case 'B':  recip= REC_BCC; break;
    }

    if (recip == -1 || (type[1] != '+' && type[1] != '\0'))
	die("Bad value for recipient type (%s)",type);

    if (addr[0] == '\0' || bad_addr(addr))
	rc= 1;
    else if (bad_header(name))
	rc= 2;
    else
    {
	email_recipient(curr_email[emid], recip, addr,
		(name[0] == '\0') ? NULL : name,
		type[1] == '+');
	rc= 0;
    }
    push_integer(rc);
    free(name);
    free(addr);
    free(type);
}


/* FUNC_EMAIL_SEND:  <subs> <contmask> <id> email_send -
 *
 *  Send the current email #<id>.  If contmask is 1, send only primary content,
 *  if it is 2, send only alternate content, otherwise send both.  Subs should
 *  be an array of strings.  If the main body text (not attachments) contains
 *  any strings like ~`19`~ (that's a number surrounded by backquotes and
 *  tildes) then the whole substring will be replaced by the 19th entry in
 *  the subs array (one-based index).  These substitutions are a bit flakey,
 *  because they happen after encoding, so (1) they won't work in anything
 *  that gets encoded in base64 (images, videos, audio, things with lots of
 *  eight bit values, things without final newlines), and (2) they could
 *  lead to usually long lines in things that get encoded as quoted-printable
 *  which may confuse some mailers (this happens if the original message
 *  contained eight bit values, internal nulls, or long lines).
 */

void func_email_send()
{
    int i, log;
    FILE *logfp= NULL;
    char **subs;

    int emid= pop_integer();
    int contmask= pop_integer();
    Array *a= pop_array();

    if (contmask < 1 || contmask > 3) contmask= 3;

    if (emid < 0 || emid >= n_email || curr_email[emid] == NULL)
	die("Email message #%d not defined",emid);

    for (i= 0; i < N_REC; i++)
    {
	if (curr_email[emid]->to[i] != NULL) break;
    }
    if (i >= N_REC)
	die("No recipients on current email message #%d",emid);

    if (curr_email[emid]->cont == NULL)
	die("No content in current email message #%d",emid);

    if (a->sz > 0)
    {
	subs= (char *)malloc(a->sz * sizeof(char *));
	for (i= 0; i < a->sz; i++)
	{
	    if (type(a->a[i]) == TK_STRING)
		subs[i]= sval(a->a[i]);
	    else
	    	die("Substution value %d in array is not a string",i+1);
	}
    }
    else
	subs= NULL;

    if ((log= idict(VAR_LOGEMAIL)) > 0 &&
    	(logfp= fopen(MAIL_LOG_FILE,"a")) != NULL)
    {
	time_t clock= time(0L);
	struct emaddr *to;
	char date[25];
	VSTRING out;
	
	if (log > 1) fputs("-----------------\n", logfp);

	sprintf(date,"%15.15s TO:", ctimez(&clock,NULL)+4);
	vs_start(&out, 200, date, strlen(date));
	for (to= curr_email[emid]->to[REC_TO]; to != NULL; to= to->next)
	{
	   vs_cat(&out, to->addr);
	   if (to->next != NULL) *vs_inc(&out)= ',';
	}
	vs_cat(&out, " SUBJECT:");
	vs_cat(&out, curr_email[emid]->subj);
	*vs_inc(&out)= '\n';
	*vs_inc(&out)= '\0';

    	fputs(out.begin,logfp);
	vs_destroy(&out);
    }

    if ((i= email_send(curr_email[emid], contmask, subs, a->sz,
    		       (log > 1) ? logfp : NULL)) > 0)
	die("Email failed:  error code %d",i);

    if (logfp != NULL)
    	fclose(logfp);
}


/* FUNC_EMAIL_CLOSE:   <id> email_close -
 *
 * Discard the current email #<id>, if there is one.
 */

void func_email_close()
{
    int emid= pop_integer();

    if (emid < 0 || emid >= n_email)
	return;

    if (curr_email[emid] != NULL)
	email_free(curr_email[emid]);
    curr_email[emid]= NULL;
}

#else

/* No-Op versions for systems without email enabled */

void func_email_new()
{
    pop_integer();
    free(pop_string());
    free(pop_string());
    free(pop_string());
    push_integer(0);
}

void func_email_body()
{
    pop_integer();
    free(pop_string());
    free(pop_string());
    free(pop_string());
    (void) pop_integer();
    push_integer(0);
}

void func_email_recip()
{
    pop_integer();
    free(pop_string());
    free(pop_string());
    free(pop_string());
    push_integer(0);
}

void func_email_send()
{
    pop_integer();
    pop_integer();
}

void func_email_close()
{
    pop_integer();
}

#endif /* EMAIL */
